from typing import Any, Collection
import pandas
import logging, pyvisa
# from ams_visa_scpi import VisaScpi
from ams_scope import ams_osciloscope
from enum import Enum
from collections import namedtuple
from time import sleep

my_logger = logging.getLogger(__name__)


class ams_agilent_scope(ams_osciloscope):
    """
    This class implements the interface that defines the osciloscope functionalities
    """

    class AcquisitionMode(Enum):
        """AcquisitionMode 
        This enum defines the acquisition modes for a scope
        """
        AUTO = 0
        CONTINUOUS = ':RUN'
        SINGLE = ':SINGLe'


    class TriggerDirection(Enum):
        """TriggerDirection 
        This enum defines the possible edge directions of a trigger edge
        """
        BOTH = 'EITH'
        RISING = 'RISING'
        FALLING = 'FALLING'


    meas_stats = namedtuple('meas_stats', 'current mean min max range std_dev')

    class TriggerSource(Enum):
        """TriggerSource 
        This enum defines the possible trigger sources
        """
        CH1 = 'CHANNEL1'
        CH2 = 'CHANNEL2'
        CH3 = 'CHANNEL3'
        CH4 = 'CHANNEL4'
        EXT = 'EXT'
        LINE = 'LINE'


    class Channels(Enum):
        """Channels 
        This enum defines the possible channels
        """
        CHANNEL_1 = 'CHANnel1'
        CHANNEL_2 = 'CHANnel2'
        CHANNEL_3 = 'CHANnel3'
        CHANNEL_4 = 'CHANnel4'


    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        Initializes a scope instrument

        Args:
            device_dict (dict): instrument settings
            example:
            {
                "Instrument_name": "my_scope",
                "Type": "ams_agilent_scope:ams_agilent_scope",
                "Details": {
                    "VISA_address" : ""
                }
            }

        Returns:
            object: switching instrument object
        """
        self.visa_address = device_dict['Details']['VISA_address']
        rm = pyvisa.ResourceManager()
        self.visaSession = rm.open_resource(self.visa_address)
        self.visaSession.timeout = 20000


    def GetInstrumentError(self) -> str:
        """GetInstrumentError 
        Gets the error string from the instrument

        Returns:
            str: error description
        """
        return self.visaSession.query('SYST:ERR?')


    def IdentifyInstrument(self) -> str:
        """IdentifyInstrument 
        Gets the identifier from the instrument

        Returns:
            str: identification string
        """
        return self.visaSession.query('*IDN?')


    def ResetInstrumehnt(self) -> None:
        """ResetInstrumehnt 
        Resets the instrument
        """
        return self.visaSession.query('*RST')


    def _checkErrorCode(self) -> bool:
        """_checkErrorCode 
        This function checks the error state of the scope

        Returns:
            bool: returns True if there is no error present in the scope
        """
        error = self.visaSession.query('SYSTem:ERRor? NUMber')
        result = int(f'{error}') == 0
        return result


    def Run(self, mode: AcquisitionMode=AcquisitionMode.AUTO) -> bool:
        """Run 
        Starts the acquisition with the scope in a certain operation mode

        Args:
            mode (AcquisitionMode, optional): operation mode of the scope. Defaults to AcquisitionMode.AUTO.

        Returns:
            bool: returns True if the command executed without any problem
        """
        if mode != self.AcquisitionMode.AUTO:
            command = mode.value
            self.visaSession.write_raw(command)
            return self._checkErrorCode()
        else:
            return True


    def Stop(self) -> bool:
        """Stop 
        Stops the acquisition

        Returns:
            bool: returns True if the command executed without any problem
        """
        command ='STOP'
        self.visaSession.write_raw(command)
        return self._checkErrorCode()


    def LoadSetup(self, setup_file_path: str) -> bool:
        """LoadSetup 
        This function loads a setup file from the scope internal memory

        Args:
            setup_file_path (str): path of the setup file

        Returns:
            bool: returns True if the command executed without any problem
        """
        command = f':DISK:LOAD "{setup_file_path}"'
        self.visaSession.write_raw(command)
        return self._checkErrorCode()
    
    
    def SaveSetup(self, setup_file_path: str) -> bool:
        """LoadSetup 
        This function loads a setup file from the scope internal memory

        Args:
            setup_file_path (str): path of the setup file

        Returns:
            bool: returns True if the command executed without any problem
        """
        command = f':DISK:SAVE:SETup "{setup_file_path}"'
        self.visaSession.write_raw(command)
        return self._checkErrorCode()


    def UploadSetupFile(self, PC_path: str, scope_path: str) -> bool:
        """UploadSetupFile 
        Uploads a file to the scopes memory from the PC

        Args:
            PC_path (str): path of the setup file on the PC
            scope_path (str): destination path of the setup file on the scope

        Returns:
            bool: returns True if the command executed without any problem
        """
        with open(PC_path, 'rb') as setup_file:
            setup_string = setup_file.read()
        to_send = self._buildIEEEblock(':SYSTem:SETup ', setup_string)
        self.visaSession.write_raw(to_send)
        sleep(1)
        self.SaveSetup(scope_path)


    def DownloadSetupFile(self, scope_path: str, PC_path: str) -> bool:
        """DownloadSetupFile 
        Downloads a setup file from scope memory to PC

        Args:
            scope_path (str): path of the setup file on the scope
            PC_path (str): target path on the PC

        Returns:
            bool: returns True if the command executed without any problem
        """
        command = f':DISK:LOAD "{scope_path}"'
        sleep(1)
        self.visaSession.write(command)
        self.visaSession.write(":SYSTem:SETup?")
        setup_raw = self.visaSession.read_raw()
        setup = self._decodeIEEEblock(setup_raw)
        with open(PC_path, 'wb') as setup_file:
            setup_file.write(setup)
        sleep(1)


    def GetScopeImage(self, PC_path: str) -> bool:
        """GetScopeImage 
        Takes a snapshot of the screen of the scope and saves it onto the PC

        Args:
            PC_path (str): image path on PC

        Returns:
            bool:  returns True if the command executed without any problem
        """
        #command = ":DISPlay:DATA? png, color"
        #command = ":DISPlay:DATA? png, screen, on, normal"
        command = ":DISPlay:DATA? png, screen, 0, normal"
        self.visaSession.write(command)
        sleep(1)
        raw_data = self.visaSession.read_raw()
        png_data = self._decodeIEEEblock(raw_data)
        with open(PC_path, 'wb') as f:
            f.write(png_data)



    def FetchData(self, channel: str=Channels.CHANNEL_1) -> list:
        """FetchData 
        Fetches the measurement data actually available on the scope

        Args:
            channel (str): channel name

        Returns:
            list: list of float type measurement values
        """
        self.visaSession.write_raw(f':WAVeform:SOURce {channel.value}')
        self.visaSession.write_raw(':WAVeform:FORMat ASCii')
        self.visaSession.write_raw(':WAVeform:DATA?')
        waveform_string = self.visaSession.read_raw().decode('utf-8').split(',')
        waveform = []
        for item in waveform_string:
            if len(item) > 2:
                waveform.append(float(item))
        return waveform


    def GetMeasurements(self, measurement_name: str, statistics: bool) -> dict:
        """GetMeasurements [summary]

        Args:
            measurement_name (str): [description]

        Returns:
            dict: [description]
        """
        result = {}
        self.scope._write(':MEASURE:RESULTS?')
        res = self.scope._read()
        result = self._parseMeasurementStats(res)
        return result


    @classmethod
    def _parseMeasurementStats(self, return_string: str) -> dict:
        """
        This method parses the string returned from the scope into a dictionary

        Args:
            return_string (str): IEEE string

        Returns:
            dict: measurement statistics parsed from the IEEE string
        """
        stat_dict = {
            "ch1": {},
            "ch2": {},
            "ch3": {},
            "ch4": {}
        }
        fields_list = return_string.split(',')
        for index in range(int(len(fields_list))):
            if '(' in fields_list[index]:
                ch_no, meas_name = self._parseMeasName(fields_list[index])
                act_ch = f'ch{ch_no}'
                stat_dict[act_ch][meas_name] = self.meas_stats(
                    fields_list[index + 1],
                    fields_list[index + 2],
                    fields_list[index + 3],
                    fields_list[index + 4],
                    fields_list[index + 5],
                    fields_list[index + 6]
                )
            else:
                pass

        return stat_dict


    @classmethod
    def _parseMeasName(self, meas_name: str) -> list:
        """
        This method extract measurement name and the channel number from the measurement name returned by the scope

        Args:
            meas_name (str): measurement name

        Returns:
            list: first element is the channel number, second element is the name of the measurement
        """
        return str(meas_name.split('(')[1])[:-1], meas_name.split('(')[0].replace(' ', '_')


    def SetTimeAxis(self, x_div_sec: float=None, x_delay_sec: float=None) -> bool:
        """
        Sets the time axis

        Args:
            x_div_sec (float): sets the div/sec scale - e.g.: write 10e-6 if you want 10us/div as scale
            x_delay_sec (float): sets the delay in seconds - e.g.: write 2e-6 if you want 2 us delay

        Returns:
            bool: returns true if no error occured
        """
        if x_div_sec != None:
            command = f':TIMEBASE:SCALE {x_div_sec:f} s'
            self.visaSession.write(command)

        if x_delay_sec != None:
            command = f':TIMEBASE:POS {x_delay_sec:f} s'
            self.visaSession.write(command)
        return self._checkErrorCode()


    def SetVertical(self, channel: int, scale: float=None, offset: float=None) -> bool:
        """
        Sets the vertical scale of the selected channel

        Args:
            channel (int): channel number
            scale (float, optional): defines the V/div value. Defaults to None.
            offset (float, optional): defines the ofset in V. Defaults to None.

        Returns:
            bool: returns true if no error occured
        """
        if scale != None:
            command = f':CHANNEL{channel}:SCALE {scale:f} V'
            self.visaSession.write(command)

        if offset != None:
            command = f':CHANNEL{channel}:OFFSET {offset:f} V'
            self.visaSession.write(command)
        return self._checkErrorCode()


    def SetTrigger(self, 
        source: TriggerSource=None, 
        edge: TriggerDirection=None, 
        level: float=None
    ) -> bool:
        """SetTrigger 
        Sets the trigger properties

        Args:
            source (TriggerSource, optional): sets the source of the trigger signal. Defaults to TriggerSource.CH1.
            edge (TriggerDirection, optional): sets the edge direction of the trigger edge. Defaults to TriggerDirection.BOTH.
            level (float, optional): sets the trigger level. Defaults to 0.

        Returns:
            bool: returns True if the command executed without any problem
        """
        if source != None:
            command = f":TRIGGER:EDGE:SOURCE {source.value}"
            self.visaSession.write(command)
        if edge != None:
            command = f":TRIGGER:EDGE:SLOPE {edge.value}"
            self.visaSession.write(command)
        if level != None:
            command = f":TRIGGER:LEVEL {source.value}, {level:f} V"
            self.visaSession.write(command)
        return self._checkErrorCode()


    @classmethod
    def _decodeIEEEblock(self, raw_data: bytes) -> str:
        """
        This method decodes the information coming from an IEEE binary block

        Args:
            raw_data (bytes): binary string

        Returns:
            str: extracted string
        """
        self._checkIfIEEE(raw_data)
        digits = self._getDataLengthDigitsNo(raw_data)
        data_length = self._getDataLength(raw_data, digits)
        start_pos = 1 + 1 + digits
        return raw_data[start_pos:]


    @classmethod
    def _checkIfIEEE(self, raw_data: bytes) -> bool:
        """
        This method checks whether the raw data is IEEE format

        Args:
            raw_data (bytes): received raw data

        Returns:
            bool: true if the received data is IEEE format
        """
        return chr(raw_data[0]) == '#'


    @classmethod
    def _getDataLengthDigitsNo(self, raw_data: bytes) -> int:
        """
        This method extract the number of digits of the data length

        Args:
            raw_data (bytes): received raw data

        Returns:
            int: number of digits of data length
        """
        return int(chr(raw_data[1]))


    @classmethod
    def _getDataLength(self, raw_data: bytes, digits: int) -> int:
        """
        Gets the data length

        Args:
            raw_data (bytes): received raw data
            digits (int): number of digits of data length

        Returns:
            int: data length
        """
        end_position = 2 + digits
        my_logger.debug(end_position)
        return int(raw_data[2:end_position].decode('utf-8'))


    @classmethod
    def _buildIEEEblock(self, header_string: str, data: bytes) -> bytes:
        block = str('#8%08d' % len(data)).encode('utf-8') + data
        to_send = header_string.encode('utf-8') + block
        return to_send