from enum import Enum
from ams_instrument.ams_instrument import ams_instrument
from abc import ABC, abstractmethod
import os, sys, pathlib
import logging
import pandas

my_logger = logging.getLogger(__name__)


class ams_osciloscope(ams_instrument):
    """
    This class implements the interface that defines the osciloscope functionalities
    """
    class AcquisitionMode(Enum):
        """AcquisitionMode 
        This enum defines the acquisition modes for a scope
        """
        AUTO = 0
        CONTINUOUS = 1
        SINGLE = 2


    class TriggerDirection(Enum):
        """TriggerDirection 
        This enum defines the possible edge directions of a trigger edge
        """
        BOTH = 0
        RISING = 1
        FALLING = 2


    class TriggerSource(Enum):
        """TriggerSource 
        This enum defines the possible trigger sources
        """
        CH1 = 0
        CH2 = 1
        CH3 = 2
        CH4 = 3
        EXT = 4
        LINE = 5


    class Channels(Enum):
        """Channels 
        This enum defines the possible channels
        """
        CHANNEL_1 = 1
        CHANNEL_2 = 2
        CHANNEL_3 = 3
        CHANNEL_4 = 3


    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        Initializes a scope instrument

        Args:
            device_dict (dict): instrument settings

        Returns:
            object: switching instrument object
        """
        pass


    @abstractmethod
    def GetInstrumentError(self) -> str:
        """GetInstrumentError 
        Gets the error string from the instrument

        Returns:
            str: error description
        """
        pass


    @abstractmethod
    def  IdentifyInstrument(self) -> str:
        """IdentifyInstrument 
        Gets the identifier from the instrument

        Returns:
            str: identification string
        """
        pass


    @abstractmethod
    def ResetInstrumehnt(self) -> None:
        """ResetInstrumehnt 
        Resets the instrument
        """
        pass


    @abstractmethod
    def Run(self, mode: AcquisitionMode=AcquisitionMode.AUTO) -> bool:
        """Run 
        Starts the acquisition with the scope in a certain operation mode

        Args:
            mode (AcquisitionMode, optional): operation mode of the scope. Defaults to AcquisitionMode.AUTO.

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def Stop(self) -> bool:
        """Stop 
        Stops the acquisition

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def LoadSetup(self, setup_file_path: str) -> bool:
        """LoadSetup 
        This function loads a setup file from the scope internal memory

        Args:
            setup_file_path (str): path of the setup file

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def SaveSetup(self, setup_file_path: str) -> bool:
        """LoadSetup 
        This function loads a setup file from the scope internal memory

        Args:
            setup_file_path (str): path of the setup file

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def UploadSetupFile(self, PC_path: str, scope_path: str) -> bool:
        """UploadSetupFile 
        Uploads a file to the scopes memory from the PC

        Args:
            PC_path (str): path of the setup file on the PC
            scope_path (str): destination path of the setup file on the scope

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def DownloadSetupFile(self, scope_path: str, PC_path: str) -> bool:
        """DownloadSetupFile 
        Downloads a setup file from scope memory to PC

        Args:
            scope_path (str): path of the setup file on the scope
            PC_path (str): target path on the PC

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def GetScopeImage(self, PC_path: str) -> bool:
        """GetScopeImage 
        Takes a snapshot of the screen of the scope and saves it onto the PC

        Args:
            PC_path (str): image path on PC

        Returns:
            bool:  returns True if the command executed without any problem
        """
        pass


    @abstractmethod
    def FetchData(self, channel: str=Channels.CHANNEL_1) -> list:
        """FetchData 
        Fetches the measurement data actually available on the scope

        Args:
            channel (str): channel name

        Returns:
            list: list of float type measurement values
        """
        pass


    @abstractmethod
    def GetMeasurements(self, measurement_name: str, statistics: bool) -> dict:
        """GetMeasurements [summary]

        Args:
            measurement_name (str): [description]

        Returns:
            dict: [description]
        """
        pass


    @abstractmethod
    def SetTimeAxis(self, x_div_sec: float=None, x_delay_sec: float=None) -> bool:
        """
        Sets the time axis

        Args:
            x_div_sec (float): sets the div/sec scale - e.g.: write 10e-6 if you want 10us/div as scale
            x_delay_sec (float): sets the delay in seconds - e.g.: write 2e-6 if you want 2 us delay

        Returns:
            bool: returns true if no error occured
        """
        pass


    @abstractmethod
    def SetVertical(self, channel: int, scale: float=None, offset: float=None) -> bool:
        """
        Sets the vertical scale of the selected channel

        Args:
            channel (int): channel number
            scale (float, optional): defines the V/div value. Defaults to None.
            offset (float, optional): defines the ofset in V. Defaults to None.

        Returns:
            bool: returns true if no error occured
        """
        pass


    @abstractmethod
    def SetTrigger(
        self, 
        source: TriggerSource=None, 
        edge: TriggerDirection=None, 
        level: float=None
    ) -> bool:
        """SetTrigger 
        Sets the trigger properties

        Args:
            source (TriggerSource, optional): sets the source of the trigger signal. Defaults to TriggerSource.CH1.
            edge (TriggerDirection, optional): sets the edge direction of the trigger edge. Defaults to TriggerDirection.BOTH.
            level (float, optional): sets the trigger level. Defaults to 0.

        Returns:
            bool: returns True if the command executed without any problem
        """
        pass
